/*	$LAAS$ */
/*
 * Copyright (c) 2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "server/jlocoHeader.h"
#include "jlocoConst.h"
#include "jlocoPrivate.h"


static BOOL transmitCheckedPosRef(POSTER_ID source,
				  GENPOS_CART_REF *destination,
				  double sdMax, double thetadMax,
				  double sddMax, double thetaddMax,
				  int *sourceNumRef,
				  int *report);


// return  0 on success
//        -1 if poster read failed
static int transmitCheckedSpeedRef(POSTER_ID source,
				   GENPOS_CART_SPEED *destination,
				   int *sourceNumRef);


/*------------------------------------------------------------------------
 * TrackStart  -  start codel of TrackStart request
 */
ACTIVITY_EVENT
trackStart(JLOCO_NAME *posterName, int *report)
{
  int res, numRef;
  
  if (transmitCheckedPosRef(jlocoPrivate->trackedPoster,
			    &jlocoPrivate->posRef,
			    jlocoPrivate->sdMax, jlocoPrivate->thetadMax,
			    jlocoPrivate->sddMax, jlocoPrivate->thetaddMax,
			    &numRef, report)) {
    jlocoPrivate->waitRefCounter = 0;
    jlocoPrivate->importantDrift = 0; // updated in cmd task
    jlocoPrivate->trackMode = JLOCO_TRACK_POSITION;
    return EXEC;    
  }

  jlocoPrivate->trackMode = JLOCO_IDLE;
  return END;    
}


/*------------------------------------------------------------------------
 * trackExec  -  permanent activity codel of TrackStart request
 */
ACTIVITY_EVENT
trackExec(JLOCO_NAME *posterName, int *report)
{
  static int prevNumRef = 0;
  int res, numRef;
  
  // check if trackMode got changed "from underneath" (ie emergency stop)
  if(JLOCO_TRACK_POSITION != jlocoPrivate->trackMode){
    PDEBUG("internal interrupt\n");
    if (EXEC_TASK_BILAN(JLOCO_CMDTASK_NUM) != OK) 
      *report = EXEC_TASK_BILAN(JLOCO_CMDTASK_NUM);
    else
      *report = S_jloco_INTERNAL_INTERRUPT;
    return END;
  }
  
  // pump data to cmd task
  if (!transmitCheckedPosRef(jlocoPrivate->trackedPoster,
			     &jlocoPrivate->posRef,
			     jlocoPrivate->sdMax, jlocoPrivate->thetadMax,
			     jlocoPrivate->sddMax, jlocoPrivate->thetaddMax,
			     &numRef, report)) {
    jlocoPrivate->trackMode = JLOCO_IDLE;
    return END;
  }
  
  // check for life watchdog
  if(SDI_F->watchdogTolerance >= 0){
    if(numRef != prevNumRef || numRef == -1)
      jlocoPrivate->waitRefCounter = 0;
    else{
      ++jlocoPrivate->waitRefCounter;
      PDEBUG("no change %d (%d)\n", jlocoPrivate->waitRefCounter, numRef);
      if(jlocoPrivate->waitRefCounter >
	 jlocoPrivate->posRef.tickPeriod + SDI_F->watchdogTolerance){
	PDEBUG("no change %d exceeds threshold %d\n",
	       jlocoPrivate->waitRefCounter,
	       jlocoPrivate->posRef.tickPeriod + SDI_F->watchdogTolerance);
	*report = S_jloco_POSTER_NOT_UPDATED;
	jlocoPrivate->trackMode = JLOCO_IDLE;
	return END; 
      }
    }
  }
  prevNumRef = numRef;
  
  return EXEC;
}


/*------------------------------------------------------------------------
 * trackStop  -  codel END of TrackStart
 */
ACTIVITY_EVENT
trackStop(JLOCO_NAME *posterName, int *report)
{
  jlocoPrivate->trackMode = JLOCO_IDLE;
  return ETHER;
}


/*------------------------------------------------------------------------
 * trackSpeedStart  -  start codel of TrackStart request
 */
ACTIVITY_EVENT
trackSpeedStart(JLOCO_NAME *posterName, int *report)
{
  int res, numRef;
  
  res = transmitCheckedSpeedRef(jlocoPrivate->trackedPoster,
				&jlocoPrivate->speedRef,
				&numRef);
  switch(res){
  case 0:
    PDEBUG("entering track speed mode\n");
    jlocoPrivate->waitRefCounter = 0;
    jlocoPrivate->trackMode = JLOCO_TRACK_SPEED;
    return EXEC;    
  case -1:
    jlocoPrivate->trackMode = JLOCO_IDLE;
    PDEBUG("poster read failed\n");
    *report = S_jloco_POSTER_READ_FAILED;
    break;
  default:
    jlocoPrivate->trackMode = JLOCO_IDLE;
    PDEBUG("transmitCheckedSpeedRef() returned %d, BUG?\n", res);
    *report = S_jloco_INTERNAL_ERROR;
  }
  return END;    
}


/*------------------------------------------------------------------------
 * trackSpeedExec  -  permanent activity codel of TrackSpeedStart request
 */
ACTIVITY_EVENT
trackSpeedExec(JLOCO_NAME *posterName, int *report)
{
  // TODO: Implement a speed ramp.
  
  static int prevNumRef = 0;
  int res, numRef;
  
  // check if trackMode got changed "from underneath" (ie emergency stop)
  if(JLOCO_TRACK_SPEED != jlocoPrivate->trackMode){
    PDEBUG("internal interrupt\n");
    if (EXEC_TASK_BILAN(JLOCO_CMDTASK_NUM) != OK) 
      *report = EXEC_TASK_BILAN(JLOCO_CMDTASK_NUM);
    else
      *report = S_jloco_INTERNAL_INTERRUPT;
    return END;
  }
  
  // pump data to cmd task
  res = transmitCheckedSpeedRef(jlocoPrivate->trackedPoster,
				&jlocoPrivate->speedRef,
				&numRef);
  switch(res){
  case 0:
    // this is what SHOULD happen
    break;
  case -1:
    jlocoPrivate->trackMode = JLOCO_IDLE;
    PDEBUG("poster read failed\n");
    *report = S_jloco_POSTER_READ_FAILED;
    return END;
  default:
    jlocoPrivate->trackMode = JLOCO_IDLE;
    PDEBUG("transmitCheckedSpeedRef() returned %d, BUG?\n", res);
    *report = S_jloco_INTERNAL_ERROR;
    return END;
  }
  
  // check for life watchdog
  if(jlocoPrivate->speedRef.v != 0 && jlocoPrivate->speedRef.w != 0 &&
     SDI_F->watchdogTolerance >= 0){
    if(numRef != prevNumRef)
      jlocoPrivate->waitRefCounter = 0;
    else{
      ++jlocoPrivate->waitRefCounter;
      PDEBUG("no change %d (%d)\n", jlocoPrivate->waitRefCounter, numRef);
      if(jlocoPrivate->waitRefCounter >
	 jlocoPrivate->speedRef.updatePeriod + SDI_F->watchdogTolerance){
	PDEBUG("no change %d exceeds threshold %d\n",
	       jlocoPrivate->waitRefCounter,
	       jlocoPrivate->speedRef.updatePeriod + SDI_F->watchdogTolerance);
	*report = S_jloco_POSTER_NOT_UPDATED;
	jlocoPrivate->trackMode = JLOCO_IDLE;
	return END;
      }
    }
  }
  prevNumRef = numRef;
  
  PVDEBUG("transmitted speed (%5.2f, %5.2f)\n",
	 jlocoPrivate->speedRef.v, jlocoPrivate->speedRef.w);
  
  return EXEC;
}


/*------------------------------------------------------------------------
 * trackSpeedStop  -  codel END of TrackSpeedStart
 */
ACTIVITY_EVENT
trackSpeedStop(JLOCO_NAME *posterName, int *report)
{
  jlocoPrivate->trackMode = JLOCO_IDLE;
  return ETHER;
}


// return  TRUE on success
//       or FALSE if poster read failed
static BOOL transmitCheckedPosRef(POSTER_ID source,
				  GENPOS_CART_REF *destination,
				  double sdMax, double thetadMax,
				  double sddMax, double thetaddMax,
				  int *sourceNumRef,
				  int *report)
{
  GENPOS_CART_REF tmpRef;
  int tmpNumRef;
  
  /* read the poster */
  if(OK != jlocoGENPOS_CART_REFPosterRead(source, &tmpRef)) {
    PDEBUG("poster read failed\n");
    *report = S_jloco_POSTER_READ_FAILED;
    return FALSE;
  }
  
  /* test if ref is ready */
  if (tmpRef.numRef == -1) {
    destination->numRef = -1;	// stay in mode, but set zero speeds
    return TRUE;
  }

  /* test values validity */
  if (tmpRef.vmax <= 0) {
    PDEBUG("reference vmax under limit\n");
    *report = S_jloco_BAD_REF;
    return FALSE;
  }
  if (tmpRef.vmax >  sdMax) {
    PDEBUG("reference vmax exceeds limit\n");
    *report = S_jloco_BAD_REF;
    return FALSE;
  }
  if (tmpRef.wmax <= 0) {
    PDEBUG("reference wmax under limit\n");
    *report = S_jloco_BAD_REF;
    return FALSE;
  }
  if (tmpRef.wmax >  thetadMax) {
    PDEBUG("reference wmax exceeds limit\n");
    *report = S_jloco_BAD_REF;
    return FALSE;
  }
  if (tmpRef.linAccelMax <= 0) {
    PDEBUG("reference linAccelMax under limit\n");
    *report = S_jloco_BAD_REF;
    return FALSE;
  }
  if (tmpRef.linAccelMax >  sddMax) {
    PDEBUG("reference linAccelMax exceeds limit\n");
    *report = S_jloco_BAD_REF;
    return FALSE;
  }
  if (tmpRef.angAccelMax <= 0) {
    PDEBUG("reference angAccelMax under limit\n");
    *report = S_jloco_BAD_REF;
    return FALSE;
  }
  if (tmpRef.angAccelMax >  thetaddMax) {
    PDEBUG("reference angAccelMax exceeds limit\n");
    *report = S_jloco_BAD_REF;
    return FALSE;
  }
  
  tmpNumRef = destination->numRef;
  memcpy(destination, &tmpRef, sizeof(tmpRef));
  if(-1 == tmpNumRef)
    destination->numRef = -1;	// stay in mode, but set zero speeds
  else
    destination->numRef = tmpNumRef == 0 ? 1 : 0;
  
  *sourceNumRef = tmpRef.numRef;
  return TRUE;
}


int transmitCheckedSpeedRef(POSTER_ID source,
			    GENPOS_CART_SPEED *destination,
			    int *sourceNumRef)
{
  GENPOS_CART_SPEED tmpSpeed;
  int tmpNumRef;
  
  if(OK != jlocoGENPOS_CART_SPEEDPosterRead(source, &tmpSpeed))
    return -1;
  
  tmpNumRef = destination->numRef;
  memcpy(destination, &tmpSpeed, sizeof(tmpSpeed));
  if(-1 == tmpNumRef)
    destination->numRef = -1;	// stay in mode, but set zero speeds
  else
    destination->numRef = tmpNumRef == 0 ? 1 : 0;
  
  *sourceNumRef = tmpSpeed.numRef;
  return 0;
}
